"""
通知頻度制御のプロパティベーステスト

hypothesis を使用して、通知頻度制御の正確性プロパティを検証します。
"""

import json
import tempfile
from datetime import datetime, timedelta
from pathlib import Path

from hypothesis import given, strategies as st
import pytest

from komon.notification import NotificationThrottle


class TestNotificationThrottleProperties:
    """NotificationThrottleクラスのプロパティベーステスト"""
    
    @given(
        metric_type=st.sampled_from(['cpu', 'memory', 'disk']),
        threshold_level=st.sampled_from(['warning', 'alert', 'critical']),
        interval_minutes=st.integers(min_value=1, max_value=120)
    )
    def test_throttle_interval_property(self, metric_type, threshold_level, interval_minutes):
        """
        P1: 通知抑制の正確性
        同一メトリクスの通知が、設定された間隔内に2回以上送信されない
        """
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.json') as f:
            history_file = Path(f.name)
        
        try:
            throttle = NotificationThrottle(
                {'enabled': True, 'interval_minutes': interval_minutes},
                history_file=history_file
            )
            
            # 1回目の通知
            should_send_1, reason_1 = throttle.should_send_notification(
                metric_type, threshold_level, 80.0
            )
            assert should_send_1 is True
            assert reason_1 == "first"
            
            throttle.record_notification(metric_type, threshold_level, 80.0)
            
            # 間隔内の2回目の通知（同じレベル）
            should_send_2, reason_2 = throttle.should_send_notification(
                metric_type, threshold_level, 81.0
            )
            assert should_send_2 is False
            assert reason_2 == "throttled"
        
        finally:
            if history_file.exists():
                history_file.unlink()
    
    @given(
        metric_type=st.sampled_from(['cpu', 'memory', 'disk']),
        previous_level=st.sampled_from(['warning', 'alert']),
        interval_minutes=st.integers(min_value=1, max_value=120)
    )
    def test_level_escalation_immediate_property(self, metric_type, previous_level, interval_minutes):
        """
        P2: 閾値レベル上昇時の即時通知
        閾値レベルが上昇した場合、通知間隔に関わらず即座に通知される
        """
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.json') as f:
            history_file = Path(f.name)
        
        try:
            throttle = NotificationThrottle(
                {'enabled': True, 'interval_minutes': interval_minutes},
                history_file=history_file
            )
            
            # 1回目の通知
            throttle.record_notification(metric_type, previous_level, 75.0)
            
            # レベルが上昇した場合（間隔内でも通知される）
            next_level = 'critical' if previous_level == 'alert' else 'alert'
            should_send, reason = throttle.should_send_notification(
                metric_type, next_level, 85.0
            )
            assert should_send is True
            assert reason == "level_up"
        
        finally:
            if history_file.exists():
                history_file.unlink()
    
    @given(
        notifications=st.lists(
            st.tuples(
                st.sampled_from(['cpu', 'memory', 'disk']),
                st.sampled_from(['warning', 'alert', 'critical']),
                st.floats(min_value=0.0, max_value=100.0)
            ),
            min_size=1,
            max_size=10
        )
    )
    def test_history_persistence_property(self, notifications):
        """
        P4: 履歴ファイルの整合性
        履歴ファイルの読み書きが正しく行われ、データが失われない
        """
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.json') as f:
            history_file = Path(f.name)
        
        try:
            throttle = NotificationThrottle(
                {'enabled': True, 'interval_minutes': 60},
                history_file=history_file
            )
            
            # 複数の通知を記録
            for metric_type, threshold_level, value in notifications:
                throttle.record_notification(metric_type, threshold_level, value)
            
            # 新しいインスタンスで履歴を読み込み
            throttle2 = NotificationThrottle(
                {'enabled': True, 'interval_minutes': 60},
                history_file=history_file
            )
            history = throttle2._load_history()
            
            # 最後の通知が正しく記録されている
            # 同じメトリクスタイプの最後の通知のみが保存される
            last_notifications = {}
            for metric_type, threshold_level, value in notifications:
                last_notifications[metric_type] = (threshold_level, value)
            
            for metric_type, (expected_level, expected_value) in last_notifications.items():
                assert metric_type in history
                assert history[metric_type]['threshold_level'] == expected_level
                # 浮動小数点の比較は近似値で
                assert abs(history[metric_type]['value'] - expected_value) < 0.01
        
        finally:
            if history_file.exists():
                history_file.unlink()
    
    @given(
        metric_type=st.sampled_from(['cpu', 'memory', 'disk']),
        threshold_level=st.sampled_from(['warning', 'alert', 'critical']),
        notification_count=st.integers(min_value=1, max_value=10)
    )
    def test_throttle_disabled_property(self, metric_type, threshold_level, notification_count):
        """
        P5: 設定無効時の動作
        頻度制御が無効の場合、全ての通知が送信される
        """
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.json') as f:
            history_file = Path(f.name)
        
        try:
            throttle = NotificationThrottle(
                {'enabled': False, 'interval_minutes': 60},
                history_file=history_file
            )
            
            for i in range(notification_count):
                should_send, reason = throttle.should_send_notification(
                    metric_type, threshold_level, 80.0 + i
                )
                assert should_send is True
                assert reason == "disabled"
                throttle.record_notification(metric_type, threshold_level, 80.0 + i)
        
        finally:
            if history_file.exists():
                history_file.unlink()



"""
3段階閾値の統合テスト

エンドツーエンドでの閾値検知と通知生成をテストします。
"""

import pytest
from komon.analyzer import load_thresholds, analyze_usage
from komon.settings_validator import ThresholdLevel


class TestThreeTierIntegration:
    """3段階閾値の統合テスト"""
    
    def test_end_to_end_three_tier_detection(self):
        """3段階閾値でのエンドツーエンド検知"""
        config = {
            "thresholds": {
                "cpu": {"warning": 70, "alert": 85, "critical": 95},
                "mem": {"warning": 70, "alert": 80, "critical": 90},
                "disk": {"warning": 70, "alert": 80, "critical": 90},
            }
        }
        
        # 設定読み込み
        thresholds = load_thresholds(config)
        
        # 使用率データ
        usage = {
            "cpu": 75.0,   # 警告レベル
            "mem": 85.0,   # 警戒レベル
            "disk": 95.0,  # 緊急レベル
        }
        
        # 分析実行
        alerts = analyze_usage(usage, thresholds)
        
        # 3つのアラートが生成される
        assert len(alerts) == 3
        
        # 各レベルの絵文字が含まれる
        alert_text = "\n".join(alerts)
        assert "💛" in alert_text  # 警告
        assert "🧡" in alert_text  # 警戒
        assert "❤️" in alert_text  # 緊急
    
    def test_backward_compatibility_with_legacy_config(self):
        """従来設定との後方互換性"""
        # 従来の単一閾値設定
        legacy_config = {
            "thresholds": {
                "cpu": 85,
                "mem": 80,
                "disk": 80,
            }
        }
        
        # 設定読み込み（3段階に正規化される）
        thresholds = load_thresholds(legacy_config)
        
        # 従来の閾値（alert）を超える値
        usage = {
            "cpu": 87.0,
            "mem": 82.0,
            "disk": 82.0,
        }
        
        # 分析実行
        alerts = analyze_usage(usage, thresholds)
        
        # すべてアラートが発生する
        assert len(alerts) == 3
        
        # 警戒レベル（alert）として検知される
        for alert in alerts:
            assert "🧡" in alert
    
    def test_escalation_scenario(self):
        """エスカレーションシナリオ"""
        config = {
            "thresholds": {
                "cpu": {"warning": 70, "alert": 85, "critical": 95},
                "mem": {"warning": 70, "alert": 80, "critical": 90},
                "disk": {"warning": 70, "alert": 80, "critical": 90},
            }
        }
        
        thresholds = load_thresholds(config)
        
        # シナリオ1: 警告レベル
        usage1 = {"cpu": 75.0, "mem": 60.0, "disk": 60.0}
        alerts1 = analyze_usage(usage1, thresholds)
        assert len(alerts1) == 1
        assert "💛" in alerts1[0]
        
        # シナリオ2: 警戒レベルにエスカレーション
        usage2 = {"cpu": 87.0, "mem": 60.0, "disk": 60.0}
        alerts2 = analyze_usage(usage2, thresholds)
        assert len(alerts2) == 1
        assert "🧡" in alerts2[0]
        
        # シナリオ3: 緊急レベルにエスカレーション
        usage3 = {"cpu": 96.0, "mem": 60.0, "disk": 60.0}
        alerts3 = analyze_usage(usage3, thresholds)
        assert len(alerts3) == 1
        assert "❤️" in alerts3[0]
    
    def test_mixed_format_config(self):
        """3段階と単一値の混在設定"""
        config = {
            "thresholds": {
                "cpu": {"warning": 70, "alert": 85, "critical": 95},
                "mem": 80,  # 単一値
                "disk": {"warning": 70, "alert": 80, "critical": 90},
            }
        }
        
        thresholds = load_thresholds(config)
        
        # すべて3段階形式に正規化される
        assert isinstance(thresholds["cpu"], dict)
        assert isinstance(thresholds["mem"], dict)
        assert isinstance(thresholds["disk"], dict)
        
        # 単一値が正しく正規化される
        assert thresholds["mem"]["alert"] == 80
        assert thresholds["mem"]["warning"] == 70  # 80 - 10
        assert thresholds["mem"]["critical"] == 90  # 80 + 10
    
    def test_notification_message_format(self):
        """通知メッセージのフォーマット確認"""
        config = {
            "thresholds": {
                "cpu": {"warning": 70, "alert": 85, "critical": 95},
                "mem": {"warning": 70, "alert": 80, "critical": 90},
                "disk": {"warning": 70, "alert": 80, "critical": 90},
            }
        }
        
        thresholds = load_thresholds(config)
        usage = {"cpu": 75.0, "mem": 60.0, "disk": 60.0}
        alerts = analyze_usage(usage, thresholds)
        
        assert len(alerts) == 1
        alert = alerts[0]
        
        # メッセージに必要な要素が含まれる
        assert "💛" in alert  # 絵文字
        assert "そろそろ気にかけておいた方がいいかも" in alert  # プレフィックス
        assert "CPU" in alert  # メトリクス名
        assert "75.0%" in alert  # 値
    
    def test_all_levels_message_content(self):
        """すべてのレベルのメッセージ内容確認"""
        config = {
            "thresholds": {
                "cpu": {"warning": 70, "alert": 85, "critical": 95},
                "mem": {"warning": 70, "alert": 80, "critical": 90},
                "disk": {"warning": 70, "alert": 80, "critical": 90},
            }
        }
        
        thresholds = load_thresholds(config)
        
        # 警告レベル
        usage_warning = {"cpu": 75.0, "mem": 60.0, "disk": 60.0}
        alerts_warning = analyze_usage(usage_warning, thresholds)
        assert "そろそろ気にかけておいた方がいいかも" in alerts_warning[0]
        
        # 警戒レベル
        usage_alert = {"cpu": 87.0, "mem": 60.0, "disk": 60.0}
        alerts_alert = analyze_usage(usage_alert, thresholds)
        assert "ちょっと気になる水準です" in alerts_alert[0]
        
        # 緊急レベル
        usage_critical = {"cpu": 96.0, "mem": 60.0, "disk": 60.0}
        alerts_critical = analyze_usage(usage_critical, thresholds)
        assert "かなり逼迫しています！" in alerts_critical[0]
    
    def test_boundary_values(self):
        """境界値でのテスト"""
        config = {
            "thresholds": {
                "cpu": {"warning": 70, "alert": 85, "critical": 95},
                "mem": {"warning": 70, "alert": 80, "critical": 90},
                "disk": {"warning": 70, "alert": 80, "critical": 90},
            }
        }
        
        thresholds = load_thresholds(config)
        
        # 閾値ちょうどの値
        usage = {"cpu": 70.0, "mem": 80.0, "disk": 90.0}
        alerts = analyze_usage(usage, thresholds)
        
        assert len(alerts) == 3
        
        # 各レベルが正しく判定される
        cpu_alert = [a for a in alerts if "CPU" in a][0]
        mem_alert = [a for a in alerts if "メモリ" in a][0]
        disk_alert = [a for a in alerts if "ディスク" in a][0]
        
        assert "💛" in cpu_alert  # warning
        assert "🧡" in mem_alert  # alert
        assert "❤️" in disk_alert  # critical
    
    def test_no_threshold_config(self):
        """閾値設定がない場合のデフォルト動作"""
        config = {}  # 空の設定
        
        thresholds = load_thresholds(config)
        
        # デフォルト値が設定される
        assert "cpu" in thresholds
        assert "mem" in thresholds
        assert "disk" in thresholds
        
        # すべて3段階形式
        assert isinstance(thresholds["cpu"], dict)
        assert "warning" in thresholds["cpu"]
        assert "alert" in thresholds["cpu"]
        assert "critical" in thresholds["cpu"]



"""
Komon - 軽量アドバイザー型SOAR風監視ツール
セットアップスクリプト
"""

from setuptools import setup, find_packages
from pathlib import Path

# README.mdを読み込む
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")

# version.txtからバージョンを読み込む
version = (this_directory / "version.txt").read_text(encoding="utf-8").strip()

setup(
    name="komon",
    version=version,
    description="軽量アドバイザー型SOAR風監視ツール",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="kamonabe",
    url="https://github.com/kamonabe/Komon",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    install_requires=[
        "psutil>=5.9.0",
        "PyYAML>=6.0",
        "requests>=2.31.0",
    ],
    entry_points={
        "console_scripts": [
            "komon=komon.cli:main",
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: POSIX :: Linux",
        "Topic :: System :: Monitoring",
    ],
    python_requires=">=3.9",
    include_package_data=True,
)




"""
多重実行プロセス検出モジュールのユニットテスト
"""

import pytest
from unittest.mock import MagicMock, patch
from komon.duplicate_detector import detect_duplicate_processes, _extract_script_name


class TestDetectDuplicateProcesses:
    """detect_duplicate_processes()のテスト"""
    
    def test_detect_duplicate_processes_normal(self):
        """
        正常系: 多重実行プロセスが検出される
        
        **検証要件: AC-001**
        """
        # モックプロセスを作成
        mock_processes = [
            {'pid': 1001, 'cmdline': ['python', '/path/to/backup.py']},
            {'pid': 1002, 'cmdline': ['python', '/path/to/backup.py']},
            {'pid': 1003, 'cmdline': ['python', '/path/to/backup.py']},
            {'pid': 1004, 'cmdline': ['python', '/path/to/backup.py']},
            {'pid': 2001, 'cmdline': ['/bin/bash', '/path/to/sync.sh']},
            {'pid': 2002, 'cmdline': ['/bin/bash', '/path/to/sync.sh']},
            {'pid': 2003, 'cmdline': ['/bin/bash', '/path/to/sync.sh']},
        ]
        
        with patch('psutil.process_iter') as mock_iter:
            mock_iter.return_value = [
                MagicMock(info=proc) for proc in mock_processes
            ]
            
            result = detect_duplicate_processes(threshold=3)
        
        # 検証
        assert len(result) == 2
        
        # backup.py
        backup = next(r for r in result if r['script'] == 'backup.py')
        assert backup['count'] == 4
        assert backup['pids'] == [1001, 1002, 1003, 1004]
        
        # sync.sh
        sync = next(r for r in result if r['script'] == 'sync.sh')
        assert sync['count'] == 3
        assert sync['pids'] == [2001, 2002, 2003]
    
    def test_detect_duplicate_processes_no_duplicates(self):
        """
        正常系: 多重実行プロセスがない場合
        
        **検証要件: AC-001**
        """
        mock_processes = [
            {'pid': 1001, 'cmdline': ['python', '/path/to/backup.py']},
            {'pid': 2001, 'cmdline': ['/bin/bash', '/path/to/sync.sh']},
        ]
        
        with patch('psutil.process_iter') as mock_iter:
            mock_iter.return_value = [
                MagicMock(info=proc) for proc in mock_processes
            ]
            
            result = detect_duplicate_processes(threshold=3)
        
        # 検証: 閾値未満なので空リスト
        assert result == []
    
    def test_detect_duplicate_processes_threshold(self):
        """
        閾値判定のテスト
        
        **検証要件: AC-002**
        """
        mock_processes = [
            {'pid': 1001, 'cmdline': ['python', '/path/to/backup.py']},
            {'pid': 1002, 'cmdline': ['python', '/path/to/backup.py']},
            {'pid': 1003, 'cmdline': ['python', '/path/to/backup.py']},
        ]
        
        with patch('psutil.process_iter') as mock_iter:
            mock_iter.return_value = [
                MagicMock(info=proc) for proc in mock_processes
            ]
            
            # 閾値3: 3個なので検出される
            result = detect_duplicate_processes(threshold=3)
            assert len(result) == 1
            assert result[0]['count'] == 3
            
            # 閾値4: 3個なので検出されない
            result = detect_duplicate_processes(threshold=4)
            assert result == []
    
    def test_detect_duplicate_processes_invalid_threshold(self):
        """
        不正な閾値の場合、デフォルト値（3）を使用
        
        **検証要件: AC-002**
        """
        mock_processes = [
            {'pid': 1001, 'cmdline': ['python', '/path/to/backup.py']},
            {'pid': 1002, 'cmdline': ['python', '/path/to/backup.py']},
            {'pid': 1003, 'cmdline': ['python', '/path/to/backup.py']},
        ]
        
        with patch('psutil.process_iter') as mock_iter:
            mock_iter.return_value = [
                MagicMock(info=proc) for proc in mock_processes
            ]
            
            # 不正な閾値（0以下）
            result = detect_duplicate_processes(threshold=0)
            assert len(result) == 1  # デフォルト値3で検出
            
            # 不正な閾値（文字列）
            result = detect_duplicate_processes(threshold="invalid")
            assert len(result) == 1  # デフォルト値3で検出
    
    def test_detect_duplicate_processes_error_handling(self):
        """
        エラーハンドリング: プロセス走査中のエラー
        
        **検証要件: AC-005**
        """
        with patch('psutil.process_iter') as mock_iter:
            # プロセス走査でエラー発生
            mock_iter.side_effect = Exception("Process iteration failed")
            
            result = detect_duplicate_processes()
        
        # 検証: エラー時は空リストを返す
        assert result == []


class TestExtractScriptName:
    """_extract_script_name()のテスト"""
    
    def test_extract_script_name_python(self):
        """
        Pythonスクリプトの抽出
        
        **検証要件: AC-001**
        """
        cmdline = ['python', '/path/to/script.py', 'arg1', 'arg2']
        result = _extract_script_name(cmdline)
        assert result == 'script.py'
    
    def test_extract_script_name_python3(self):
        """
        Python3スクリプトの抽出
        
        **検証要件: AC-001**
        """
        cmdline = ['python3', '/home/user/backup.py']
        result = _extract_script_name(cmdline)
        assert result == 'backup.py'
    
    def test_extract_script_name_shell(self):
        """
        シェルスクリプトの抽出
        
        **検証要件: AC-001**
        """
        cmdline = ['/bin/bash', '/path/to/sync.sh']
        result = _extract_script_name(cmdline)
        assert result == 'sync.sh'
    
    def test_extract_script_name_ruby(self):
        """
        Rubyスクリプトの抽出
        
        **検証要件: AC-001**
        """
        cmdline = ['ruby', '/path/to/script.rb']
        result = _extract_script_name(cmdline)
        assert result == 'script.rb'
    
    def test_extract_script_name_perl(self):
        """
        Perlスクリプトの抽出
        
        **検証要件: AC-001**
        """
        cmdline = ['perl', '/path/to/script.pl']
        result = _extract_script_name(cmdline)
        assert result == 'script.pl'
    
    def test_extract_script_name_no_script(self):
        """
        スクリプトがない場合
        
        **検証要件: AC-001**
        """
        cmdline = ['ls', '-la', '/tmp']
        result = _extract_script_name(cmdline)
        assert result is None
    
    def test_extract_script_name_module_execution(self):
        """
        モジュール実行（python -m module）は対象外
        
        **検証要件: AC-001**
        """
        cmdline = ['python', '-m', 'http.server', '8000']
        result = _extract_script_name(cmdline)
        assert result is None
    
    def test_extract_script_name_empty_cmdline(self):
        """
        空のコマンドライン
        
        **検証要件: AC-001**
        """
        result = _extract_script_name([])
        assert result is None
    
    def test_extract_script_name_none_cmdline(self):
        """
        Noneのコマンドライン
        
        **検証要件: AC-001**
        """
        result = _extract_script_name(None)
        assert result is None
    
    def test_extract_script_name_unsupported_extension(self):
        """
        対象外の拡張子
        
        **検証要件: AC-001**
        """
        cmdline = ['java', '-jar', '/path/to/app.jar']
        result = _extract_script_name(cmdline)
        assert result is None
